import torch


# multi-task heteroscedastic regression loss
def regression_loss(y_true, y_pred):

    y_true = y_true.reshape(-1,1)
    mu = y_pred[:, :1] # first output neuron: predicted expression value
    log_sig = y_pred[:, 1:] # second output neuron: uncertainty estimation
    sig = torch.exp(log_sig) 

    return torch.mean(2*log_sig + ((y_true-mu)/sig)**2) 





import numpy as np
import torch
from numpy import linalg as LA


def get_proto_norm(feature, centroid, ps_label, num_protos):
    num_data = feature.shape[0]
    each_cluster_num = np.zeros([num_protos])

    for i in range(num_protos):
        each_cluster_num[i] = np.sum(ps_label==i)
    
    proto_norm_term = np.zeros([num_protos])
    for i in range(num_protos):
        norm_sum = 0
        for j in range(num_data):
            if ps_label[j] == i:
                norm_sum = norm_sum + LA.norm(feature[j] - centroid[i], 2)
        proto_norm_term[i] = norm_sum / (each_cluster_num[i] * np.log2(each_cluster_num[i] + 10))
    proto_norm = torch.Tensor(proto_norm_term)
    return proto_norm



def get_proto_loss(feature, centroid, ps_label, proto_norm):

    feature_norm = torch.norm(feature, dim=-1)
    feature = torch.div(feature, feature_norm.unsqueeze(1)) 
    centroid_norm = torch.norm(centroid, dim=-1)
    centroid = torch.div(centroid, centroid_norm.unsqueeze(1))
    sim_zc = torch.matmul(feature, centroid.t())   
    sim_zc_normalized = torch.exp(sim_zc)
    sim_2centroid = torch.gather(sim_zc_normalized, -1, ps_label) 
    sim_sum = torch.sum(sim_zc_normalized, -1, keepdim=True)
    sim_2centroid = torch.div(sim_2centroid, sim_sum)
    loss = torch.mean(sim_2centroid.log())
    loss = -1 * loss
    return loss




import torch
import torch.nn as nn
# from peptide_uncertainty_utils import *
# from . import peptide_uncertainty_utils 
from .peptide_uncertainty_utils import *
import random
import torch.utils.data as Data
from .multi_task_heteroscedastic_regression_loss import regression_loss


class peptide_CNN(nn.Module):
    def __init__(self, num_amino_acid, max_pool_size, hidden_dim, output_dim, conv_layers, dropout_rate, kernel_nums, kernel_size):
        super(peptide_CNN,self).__init__()
        self.conv_layers = conv_layers
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.num_amino_acid = num_amino_acid
        self.kernel_nums = kernel_nums
        self.kernel_size = kernel_size
        self.max_pool_size = max_pool_size
        self.convs = []
        for i in range(self.conv_layers):
            if i == 0:
                self.base_conv = nn.Sequential(nn.Conv1d(in_channels=self.num_amino_acid, 
                                                        out_channels=self.kernel_nums[i], 
                                                        kernel_size = self.kernel_size[i]),
                                                        nn.BatchNorm1d(num_features=self.kernel_nums[i]), 
                                                        nn.ReLU(),
                                                        nn.MaxPool1d(kernel_size=self.max_pool_size))
            else:
                self.base_conv = nn.Sequential(nn.Conv1d(in_channels=self.kernel_nums[i-1], 
                                                        out_channels=self.kernel_nums[i], 
                                                        kernel_size = self.kernel_size[i]),
                                                        nn.BatchNorm1d(num_features=self.kernel_nums[i]), 
                                                        nn.ReLU(),
                                                        nn.MaxPool1d(kernel_size=self.max_pool_size))
            self.convs.append(self.base_conv)
        self.convs = nn.ModuleList(self.convs)
        self.fc1 = nn.Linear(3100,self.hidden_dim)

        self.fc2 = nn.Linear(self.hidden_dim,self.output_dim)
        self.dropout = nn.Dropout(dropout_rate)

    def forward(self,x):
        for i in range(self.conv_layers):
            x = self.convs[i](x)
        
        x = torch.flatten(x, start_dim=1, end_dim=-1)
        x = self.dropout(x)
        x = self.fc1(x)
        x = self.fc2(x)
        return x



class scPROTEIN_stage1_learning(nn.Module):
    def __init__(self, model, peptide_onehot_padding, Y_label,learning_rate, weight_decay, split_percentage, num_epochs,batch_size):
        super(scPROTEIN_stage1_learning, self).__init__()
        self.model = model
        self.peptide_onehot_padding = peptide_onehot_padding
        self.Y_label = Y_label
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.split_percentage = split_percentage
        self.num_epochs = num_epochs
        self.batch_size = batch_size


    def transform_peptide_data_into_dataloader(self):

        indices = list(range(self.peptide_onehot_padding.shape[0]))
        random.shuffle(indices)
        train_size = int((self.split_percentage) * len(indices))
        train_indices = indices[:train_size]
        valid_indices = indices[train_size:]
        x_train, x_valid = self.peptide_onehot_padding[train_indices], self.peptide_onehot_padding[valid_indices]
        y_train, y_valid = self.Y_label[train_indices], self.Y_label[valid_indices]
        train_dataset_split = Data.TensorDataset(x_train, y_train)
        loader = Data.DataLoader(dataset=train_dataset_split,batch_size=self.batch_size,shuffle=True)
        return loader

    def train(self):
        num_cells = self.Y_label.shape[1]
        loader = self.transform_peptide_data_into_dataloader()
        criterion = nn.MSELoss() 
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.learning_rate, weight_decay=self.weight_decay, eps=1e-3)
        for epoch in range(self.num_epochs):
            self.model.train()
            loss_all = 100.
            for step, (batch_x, batch_y) in enumerate(loader):
                loss = 0.
                optimizer.zero_grad()
                y_predict = self.model(batch_x)

                for i in range(0,2*num_cells,2):
                    loss_cell = regression_loss(batch_y[:,int(i/2)], y_predict[:,i:i+2])
                    loss += loss_cell
                loss = loss/num_cells

                loss.backward()
                optimizer.step()
                loss_all += loss.item()

            self.model.eval()
            print('epoch {}, loss_regression: {}'.format(epoch, loss_all))

        # torch.save(self.model.state_dict(), 'scPROTEIN_stage1.pkl')   


    def uncertainty_generation(self):
        self.model.eval()
        y_predict_all = self.model(self.peptide_onehot_padding)
        y_predict_all = y_predict_all.cpu().detach().numpy()

        log_uncertainty = y_predict_all[:,1::2]
        uncertainty = np.exp(log_uncertainty)
        return uncertainty
        # np.save('peptide_uncertainty_tutorial.npy',uncertainty)



from setuptools import setup, find_packages


setup(
    name='scprotein', 
    version='1.1.1',  
    author='LiWei', 
    description='scprotein is a deep contrastive learning framework for single-cell proteomics embedding.',  
    url='https://github.com/TencentAILabHealthcare/scPROTEIN',
    packages=find_packages(),  
    install_requires=[  
        'torch>=1.10.0',
        'scanpy>=1.8.2',
        'numpy>=1.22.3',
        'pandas>=1.3.5',
        'scipy>=1.8.1',
        'scikit_learn>=1.1.1',
        'torch_geometric>=2.0.4'
    ],
    python_requires=">=3.8",
    package_data = {'': ['*.txt', '*.rst',  '*.jpg']},
    author_email = 'nkuweili@mail.nankai.edu.cn',
    license='Apache'
    
)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import GCNConv
from torch_geometric.utils import dropout_adj
from .prototype_loss import *
from sklearn.cluster import KMeans
from .utils import *

'''
part of code is borrowed from https://github.com/CRIPAC-DIG/GRACE
'''


# mask feature function
def drop_feature(x, drop_prob):
    drop_mask = torch.empty(
        (x.size(1), ),
        dtype=torch.float32,
        device=x.device).uniform_(0, 1) < drop_prob
    x = x.clone()
    x[:, drop_mask] = 0

    return x

# GCN encoder
class Encoder(torch.nn.Module):
    def __init__(self, in_channels: int, out_channels: int, activation,
                 base_model=GCNConv, k: int = 2):
        super(Encoder, self).__init__()
        self.base_model = base_model

        assert k >= 2
        self.k = k
        self.conv = [base_model(in_channels, 2 * out_channels)]
        for _ in range(1, k-1):
            self.conv.append(base_model(2 * out_channels, 2 * out_channels))
        self.conv.append(base_model(2 * out_channels, out_channels))
        self.conv = nn.ModuleList(self.conv)

        self.activation = activation

    def forward(self, x: torch.Tensor, edge_index: torch.Tensor):
        for i in range(self.k):
            # print(self.conv[i](x, edge_index))
            x = self.activation(self.conv[i](x, edge_index))
            # print(x)
        return x


class Model(torch.nn.Module):
    def __init__(self, encoder: Encoder, num_hidden: int, num_proj_hidden: int,
                 tau: float = 0.5):
        super(Model, self).__init__()
        self.encoder: Encoder = encoder
        self.tau: float = tau

        self.fc1 = torch.nn.Linear(num_hidden, num_proj_hidden)
        self.fc2 = torch.nn.Linear(num_proj_hidden, num_hidden)

    def forward(self, x: torch.Tensor,
                edge_index: torch.Tensor) -> torch.Tensor:
        return self.encoder(x, edge_index)

    def projection(self, z: torch.Tensor) -> torch.Tensor:
        z = F.elu(self.fc1(z))
        return self.fc2(z)

    def sim(self, z1: torch.Tensor, z2: torch.Tensor):

        z1 = F.normalize(z1)
        z2 = F.normalize(z2)
        return torch.mm(z1, z2.t())

    def semi_loss(self, z1: torch.Tensor, z2: torch.Tensor):
        f = lambda x: torch.exp(x / self.tau)
        refl_sim = f(self.sim(z1, z1))
        between_sim = f(self.sim(z1, z2))
        return -torch.log(
            between_sim.diag()
            / (refl_sim.sum(1) + between_sim.sum(1) - refl_sim.diag()))


    def batched_semi_loss(self, z1: torch.Tensor, z2: torch.Tensor,
                          batch_size: int):
        device = z1.device
        num_nodes = z1.size(0)
        num_batches = (num_nodes - 1) // batch_size + 1
        f = lambda x: torch.exp(x / self.tau)
        indices = torch.arange(0, num_nodes).to(device)
        losses = []

        for i in range(num_batches):
            mask = indices[i * batch_size:(i + 1) * batch_size]
            refl_sim = f(self.sim(z1[mask], z1))  # [B, N]
            between_sim = f(self.sim(z1[mask], z2))  # [B, N]

            losses.append(-torch.log(
                between_sim[:, i * batch_size:(i + 1) * batch_size].diag()
                / (refl_sim.sum(1) + between_sim.sum(1)
                   - refl_sim[:, i * batch_size:(i + 1) * batch_size].diag())))

        return torch.cat(losses)

    # loss definition
    def loss(self, z1: torch.Tensor, z2: torch.Tensor,
             mean: bool = True, batch_size: int = 0):
        
        h1 = self.projection(z1)
        h2 = self.projection(z2)

        if batch_size == 0:
            l1 = self.semi_loss(h1, h2)
            l2 = self.semi_loss(h2, h1)
        else:
            l1 = self.batched_semi_loss(h1, h2, batch_size)
            l2 = self.batched_semi_loss(h2, h1, batch_size)

        ret = (l1 + l2) * 0.5
        ret = ret.mean() if mean else ret.sum()

        return ret



class scPROTEIN_learning(torch.nn.Module):

    def __init__(self, model,device, data, drop_feature_rate_1,drop_feature_rate_2,drop_edge_rate_1,drop_edge_rate_2,
                 learning_rate, weight_decay, num_protos, topology_denoising, num_epochs, alpha, num_changed_edges, seed):
        super(scPROTEIN_learning, self).__init__()
        self.model = model
        self.data = data
        self.device = device
        self.drop_feature_rate_1 = drop_feature_rate_1
        self.drop_feature_rate_2 = drop_feature_rate_2
        self.drop_edge_rate_1 = drop_edge_rate_1
        self.drop_edge_rate_2 = drop_edge_rate_2
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.num_protos = num_protos
        self.topology_denoising = topology_denoising
        self.num_epochs = num_epochs
        self.alpha = alpha
        self.seed = seed
        self.num_changed_edges = num_changed_edges
        
    def train(self):
        setup_seed(self.seed)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.learning_rate, weight_decay=self.weight_decay)
        for epoch in range(1, self.num_epochs + 1):
            self.model.train()
            
            optimizer.zero_grad()
            edge_index_1 = dropout_adj(self.data.edge_index, p=self.drop_edge_rate_1)[0]
            edge_index_2 = dropout_adj(self.data.edge_index, p=self.drop_edge_rate_2)[0]
            x_1 = drop_feature(self.data.x, self.drop_feature_rate_1)
            x_2 = drop_feature(self.data.x, self.drop_feature_rate_2)
            z1 = self.model(x_1, edge_index_1)
            z2 = self.model(x_2, edge_index_2)
            loss_node = self.model.loss(z1, z2, batch_size=0)
            # return loss



            # embedding = test(model, data.x.to(device), data.edge_index.to(device))
            embedding = self.test()
            embedding_cpu = embedding.cpu().detach().numpy()
            kmeans = KMeans(n_clusters=self.num_protos).fit(embedding_cpu)
            label_kmeans = kmeans.labels_
            centers = np.array([np.mean(embedding_cpu[label_kmeans == i,:], axis=0)
                                for i in range(self.num_protos)])
            label_kmeans = label_kmeans[:, np.newaxis]
            proto_norm = get_proto_norm(embedding_cpu, centers,label_kmeans,self.num_protos)
            centers = torch.Tensor(centers).to(self.device)
            label_kmeans = torch.Tensor(label_kmeans).long().to(self.device)
            proto_norm = torch.Tensor(proto_norm).to(self.device)
            loss_proto = get_proto_loss(embedding, centers, label_kmeans, proto_norm)       



            # topology denoising
            if self.topology_denoising:
                with torch.no_grad():
                    embedding_cpu = sp.coo_matrix(embedding_cpu)
                    similarity_matrix = embedding_cpu.dot(embedding_cpu.transpose())
                    similarity_matrix = similarity_matrix.tocoo()
                    coords = list(np.vstack((similarity_matrix.row, similarity_matrix.col)).transpose())
                    coords_tuple = []
                    for i in coords:
                        coords_tuple.append(tuple(i))

                    simi_data = list(similarity_matrix.data)
                    coord_value_dict = {}
                    for i in range(len(coords_tuple)):
                        coord_value_dict[coords_tuple[i]] = simi_data[i]

                    coord_value_dict_wo_diag = {}
                    cnt = 0
                    for key,value in coord_value_dict.items():
                        if key[0] == key[1]:
                            cnt += 1
                        else:
                            coord_value_dict_wo_diag[key] = value

                    coords_wo_diag = list(coord_value_dict_wo_diag.keys())
                    simi_data_wo_diag = np.array(list(coord_value_dict_wo_diag.values()))
                    simi_sort = simi_data_wo_diag.argsort()
                    high_prob_indices = list(simi_sort[-self.num_changed_edges:])
                    low_prob_indices = list(simi_sort[:self.num_changed_edges])
                    high_prob_coords = []
                    low_prob_coords = []
                    for i in high_prob_indices:
                        high_prob_coords.append(coords_wo_diag[i])
                    for i in low_prob_indices:
                        low_prob_coords.append(coords_wo_diag[i])

                    edge_index_now = list(self.data.edge_index.cpu().detach().numpy().T)
                    edge_index_now_list = []
                    for i in edge_index_now:
                        edge_index_now_list.append(tuple(i))
                    cnt_add = 0
                    for i in high_prob_coords:
                        if i not in edge_index_now_list:
                            edge_index_now_list.append(i)
                            cnt_add += 1
                    cnt_remove = 0
                    for i in low_prob_coords:
                        if i in edge_index_now_list:
                            edge_index_now_list.remove(i)
                            cnt_remove += 1

                    edge_index = torch.tensor(np.array(edge_index_now_list).T, dtype=torch.long).to(self.device)
                    self.data.edge_index = edge_index    
            
            
            loss = loss_node + self.alpha*loss_proto
            loss.backward()
            optimizer.step()

            print(f'(T) | Epoch={epoch:03d}, loss={loss:.4f} ')


    def test(self):
        self.model.eval()
        z = self.model(self.data.x, self.data.edge_index)
        return z

    def embedding_generation(self):
        self.model.eval()
        z = self.model(self.data.x, self.data.edge_index)
        return z.cpu().detach().numpy()










import os
import numpy as np
import pandas as pd
import torch
import torch.optim.lr_scheduler as lr_scheduler
from torch_geometric.data import Data
import scipy.sparse as sp
import random
import scanpy as sc
from operator import itemgetter
import pkg_resources

def load_sc_proteomic_features(stage1):

    # -----for data provided with peptide-level data, scPROTEIN starts from stage 1-----
    if stage1:
        feature_file = pd.read_csv('./data/Peptides-raw.csv')

        if os.path.exists('./peptide_uncertainty_estimation/peptide_uncertainty.npy'):
            peptide_uncertainty = np.load('./peptide_uncertainty_estimation/peptide_uncertainty.npy')

            

        else:
            raise Exception('-----To start from stage 1, you have to run peptide_uncertainty.py in the folder peptide_uncertainty_estimation to obtain the estimated peptide uncertainty first-----')

        cell_list = feature_file.columns.tolist()[2:]
        feature_fill = feature_file.fillna(0.)
        proteins_all = list(pd.unique(feature_fill['protein']))
        features_all = feature_fill.values[:,2:]
        weighted_protein_feature_all = []



        # obtain the uncertainty-guided protein-level data
        for i in proteins_all:
            protein_index = feature_fill[(feature_fill.protein==i)].index.tolist()
            peptide_uncertainty_subset = peptide_uncertainty[protein_index,:]
            peptide_scores = 1./peptide_uncertainty_subset
            peptide_features_subset = features_all[protein_index,:]

            weighted_peptide_features_subset = np.multiply(peptide_features_subset, peptide_scores)
            weighted_protein_feature = np.sum(weighted_peptide_features_subset, axis=0)
            weighted_protein_feature_all.append(weighted_protein_feature)
        
        features = np.array(weighted_protein_feature_all)

    # -----for data provided directly from protein-level data, scPROTEIN starts from stage 2-----
    else:
        feature_file = pd.read_csv('./data/Peptides-raw.csv')
        cell_list = feature_file.columns.tolist()[2:]
        feature_fill = feature_file.fillna(0.)
        features = feature_fill.groupby('protein').sum()
        proteins_all = list(features.index)
        features = np.array(features.values)

    return proteins_all, cell_list, features.T


def graph_generation(features, threshold, feature_preprocess):
    # take cell*protein matrix as input
    # output cell*cell graph
    features = features.astype(float)  
    features_pd = pd.DataFrame(features.T)
    adj = features_pd.corr()
    adj_matrix = np.where(adj>threshold,1,0)

    if feature_preprocess:
        features = sp.coo_matrix(features)
        features, _ = preprocess_features(features)

    adj_matrix_sp = sp.coo_matrix(adj_matrix)
    edge_index = torch.tensor(np.vstack((adj_matrix_sp.row, adj_matrix_sp.col)), dtype=torch.long)
    features = torch.tensor(features,dtype=torch.float32)

    data = Data(x=features, edge_index=edge_index)
    return data


def preprocess_graph(adj):

    adj = sp.coo_matrix(adj)
    adj_ = adj + sp.eye(adj.shape[0])
    rowsum = np.array(adj_.sum(1))
    degree_mat_inv_sqrt = sp.diags(np.power(rowsum, -0.5).flatten())
    adj_normalized = adj_.dot(degree_mat_inv_sqrt).transpose().dot(degree_mat_inv_sqrt).tocoo()
    return adj_normalized, sparse_to_tuple(adj_normalized)


def sparse_to_tuple(sparse_mx):

    if not sp.isspmatrix_coo(sparse_mx):
        sparse_mx = sparse_mx.tocoo()
    coords = np.vstack((sparse_mx.row, sparse_mx.col)).transpose()
    values = sparse_mx.data
    shape = sparse_mx.shape
    return coords, values, shape


def load_cell_type_labels():
    cell_type_file = pd.read_csv('./data/Cells.csv')
    labels = list(cell_type_file.iloc[0,1:].values)
    return labels





# load overlap protein data from two datasets
def integrate_sc_proteomic_features(dataset1, dataset2):
    
    # load individual scp data
    adata1 = sc.read_h5ad('./integration_dataset/{}.h5ad'.format(dataset1))
    adata2 = sc.read_h5ad('./integration_dataset/{}.h5ad'.format(dataset2))
    protein_data1, protein_data2 = adata1.X, adata2.X
    protein_data1 = np.nan_to_num(protein_data1)
    protein_data2 = np.nan_to_num(protein_data2)
    cell_num1, cell_num2 = protein_data1.shape[0], protein_data2.shape[0]
    proteins1, proteins2 = list(adata1.var_names),list(adata2.var_names)
    
    # define batch label and cell type labels for both two datasets
    batch_label = np.concatenate((np.zeros(cell_num1),np.ones(cell_num2))).astype(int)
    cell_type1,cell_type2 = list(adata1.obs['cell_type']), list(adata2.obs['cell_type'])
    cell_type_with_dataname = cell_type1+cell_type2
    
    cell_type1 = [i.split('(')[0] for i in cell_type1]
    cell_type2 = [i.split('(')[0] for i in cell_type2]
    overlap_cell_type = list(set(cell_type1) & set(cell_type2))
    print('overlap celltype:',overlap_cell_type)
    
    cell_type_all = cell_type1+cell_type2
    cell_type_dic = dict(zip(set(cell_type_all), range(len(set(cell_type_all)))))
    cell_type_label = np.array(itemgetter(*list(cell_type_all))(cell_type_dic))
    overlap_cell_type_label = [cell_type_dic[i] for i in overlap_cell_type]
    
    
    # search overlap protein from both two datasets
    proteins1_pd,proteins2_pd = pd.DataFrame(proteins1,columns=['protein_name']),pd.DataFrame(proteins2,columns=['protein_name'])
    overlap_protein = pd.merge(proteins1_pd, proteins2_pd, on=['protein_name'])
    overlap_protein = list(overlap_protein['protein_name'])
    print('overlap protein nums:',len(overlap_protein))
                           

    # construct overlap protein features
    features_concat = np.zeros((cell_num1+cell_num2,len(overlap_protein)))
    for i,protein in enumerate(overlap_protein):
        index1 = proteins1.index(protein)
        protein_data1_slice = protein_data1[:,index1]
        index2 = proteins2.index(protein)
        protein_data2_slice = protein_data2[:,index2]
        protein_data_slice_concat = np.concatenate([protein_data1_slice,protein_data2_slice])
        features_concat[:,i] = protein_data_slice_concat    
    
    return batch_label,cell_type_with_dataname,cell_type_label,overlap_cell_type_label, features_concat

def setup_seed(seed):
    #--- Fix random seed ---#
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.cuda.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False 
    os.environ['PYTHONHASHSEED'] = str(seed)


def preprocess_features(features):
    """Row-normalize feature matrix and convert to tuple representation"""
    rowsum = np.array(features.sum(1))
    r_inv = np.power(rowsum, -1).flatten()
    r_inv[np.isinf(r_inv)] = 0.
    r_mat_inv = sp.diags(r_inv)
    features = r_mat_inv.dot(features)
    return features.todense(), sparse_to_tuple(features)




import numpy as np
import pandas as pd
import os 
import argparse
import os.path as osp
import random
from time import perf_counter as t
import torch
import torch_geometric.transforms as T
import torch.nn.functional as F
import torch.nn as nn
import numpy as np 
import scipy.sparse as sp
from numpy import linalg as LA
from torch.nn.utils.rnn import pad_sequence
import re
from operator import itemgetter



# load data
def load_peptide(file_path):

    # load raw peptide data
    data_path = os.path.join(file_path)
    # data_path = os.path.join(os.path.abspath('..'),'data/Peptides-raw.csv')
    feature_file = pd.read_csv(data_path)
    feature_fill = feature_file.fillna(0.)
    peptides = list(feature_fill['peptide'])
    proteins = list(feature_fill['protein'])
    expression_data = np.array(feature_fill.iloc[:,2:])
    cell_list = feature_file.columns.tolist()[2:]
    num_cells = len(cell_list)
    print('peptides nums in total:', len(peptides))
    print('cell nums:',num_cells)
    return peptides, proteins, torch.HalfTensor(expression_data), cell_list,num_cells


# extract peptide sequence
def extract_peptide_seq(peptides):
    peptide_list = []
    peptide_list_charge = []
    for i in peptides:
        seq = re.sub(r'\(.*?\)','',i).replace(')', '')
        seq_split = seq.split('_')
        # print(seq_split)
        peptide_list.append(seq_split[1])
        peptide_list_charge.append(seq_split[1]+'_'+seq_split[2])

    return peptide_list,peptide_list_charge


# # extract amino acid set
# def extract_amino_acid_set(peptide_list):
#     amino_acid_set = set()
#     max_length = 0
#     for i in peptide_list:
#         seq_set = set(list(i))
#         amino_acid_set = amino_acid_set.union(seq_set)
#         seq_length = len(list(i))
#         if seq_length>max_length:
#             max_length = seq_length

#     return list(amino_acid_set),max_length


# one hot encoding
def one_hot_encode(peptide_list,amino_acid_dict):
    num_amino_acid = len(amino_acid_dict)  
    peptide_onehot_list = []
    for i in peptide_list:
        peptide_tensor = torch.tensor(itemgetter(*list(i))(amino_acid_dict))
        peptide_onehot = F.one_hot(peptide_tensor,num_classes = num_amino_acid)
        peptide_onehot_list.append(peptide_onehot)
    peptide_onehot_list_padding = pad_sequence(peptide_onehot_list,batch_first = True).permute(0,2,1)
    return peptide_onehot_list_padding.float(),num_amino_acid


def peptide_encode(peptides):

    peptide_list, peptide_list_charge = extract_peptide_seq(peptides)   
    
    # fixed dic
    amino_acid_dict = {'N': 0, 'M': 1, 'K': 2, 'W': 3, 'V': 4, 'C': 5, 'R': 6, 'Q': 7, 'G': 8, 'F': 9, 
                       'P': 10, 'S': 11, 'H': 12, 'Y': 13, 'D': 14, 'T': 15, 'L': 16, 'E': 17, 'A': 18, 'I': 19}

    peptide_onehot_list_padding,num_amino_acid = one_hot_encode(peptide_list,amino_acid_dict)
    return peptide_onehot_list_padding, num_amino_acid




# load batch label
def load_cell_type_labels():
    cell_type_file = pd.read_csv('Cells.csv')
    # print(cell_type_file)
    labels = list(cell_type_file.iloc[0,1:].values)
    batch_labels = list(cell_type_file.iloc[3,1:].values)
    return labels,batch_labels



def setup_seed(seed):
    #--- Fix random seed ---#
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.cuda.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False 
    os.environ['PYTHONHASHSEED'] = str(seed)